@@ -1,3 +1,5 @@
+"""Module that contains the command line application."""
+
 # Why does this file exist, and why not put this in `__main__`?
 #
 # You might be tempted to import things from `__main__` later,
@@ -9,37 +11,48 @@
 # - When you import `__main__` it will get executed again (as a module) because
 #   there's no `griffe.__main__` in `sys.modules`.
 
-"""Module that contains the command line application."""
-
 from __future__ import annotations
 
 import argparse
 import json
 import logging
 import os
-import subprocess  # noqa: S404
 import sys
-from datetime import datetime
+from datetime import datetime, timezone
 from pathlib import Path
-from typing import IO, Any, Callable, Sequence, Type
+from typing import IO, TYPE_CHECKING, Any, Callable, Sequence
 
 import colorama
 
-from griffe.agents.extensions import Extension, Extensions
-from griffe.agents.extensions.base import load_extensions
+from griffe import debug
 from griffe.diff import ExplanationStyle, find_breaking_changes
 from griffe.docstrings.parsers import Parser
 from griffe.encoders import JSONEncoder
-from griffe.exceptions import ExtensionError
-from griffe.git import load_git
-from griffe.loader import GriffeLoader, load  # noqa: WPS347
+from griffe.exceptions import ExtensionError, GitError
+from griffe.extensions.base import load_extensions
+from griffe.git import _get_latest_tag, _get_repo_root, load_git
+from griffe.loader import GriffeLoader, load
 from griffe.logger import get_logger
+from griffe.stats import _format_stats
+
+if TYPE_CHECKING:
+    from griffe.extensions import Extensions, ExtensionType
+
 
 DEFAULT_LOG_LEVEL = os.getenv("GRIFFE_LOG_LEVEL", "INFO").upper()
 logger = get_logger(__name__)
 
 
-def _print_data(data: str, output_file: str | IO | None):
+class _DebugInfo(argparse.Action):
+    def __init__(self, nargs: int | str | None = 0, **kwargs: Any) -> None:
+        super().__init__(nargs=nargs, **kwargs)
+
+    def __call__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
+        debug.print_debug_info()
+        sys.exit(0)
+
+
+def _print_data(data: str, output_file: str | IO | None) -> None:
     if isinstance(output_file, str):
         with open(output_file, "w") as fd:
             print(data, file=fd)
@@ -49,123 +62,45 @@ def _print_data(data: str, output_file: str | IO | None):
         print(data, file=output_file)
 
 
-def _get_latest_tag(path: str | Path) -> str:
-    if isinstance(path, str):
-        path = Path(path)
-    if not path.is_dir():
-        path = path.parent
-    output = subprocess.check_output(  # noqa: S603,S607
-        ["git", "describe", "--tags", "--abbrev=0"],
-        cwd=path,
-    )
-    return output.decode().strip()
-
-
-def _get_repo_root(path: str | Path) -> str:
-    if isinstance(path, str):
-        path = Path(path)
-    if not path.is_dir():
-        path = path.parent
-    output = subprocess.check_output(  # noqa: S603,S607
-        ["git", "rev-parse", "--show-toplevel"],
-        cwd=path,
-    )
-    return output.decode().strip()
-
-
-def _stats(stats):
-    lines = []
-    packages = stats["packages"]
-    modules = stats["modules"]
-    classes = stats["classes"]
-    functions = stats["functions"]
-    attributes = stats["attributes"]
-    objects = sum((modules, classes, functions, attributes))
-    lines.append("Statistics")
-    lines.append("---------------------")
-    lines.append("Number of loaded objects")
-    lines.append(f"  Modules: {modules}")
-    lines.append(f"  Classes: {classes}")
-    lines.append(f"  Functions: {functions}")
-    lines.append(f"  Attributes: {attributes}")
-    lines.append(f"  Total: {objects} across {packages} packages")
-    per_ext = stats["modules_by_extension"]
-    builtin = per_ext[""]
-    regular = per_ext[".py"]
-    stubs = per_ext[".pyi"]
-    compiled = modules - builtin - regular - stubs
-    lines.append("")
-    lines.append(f"Total number of lines: {stats['lines']}")
-    lines.append("")
-    lines.append("Modules")
-    lines.append(f"  Builtin: {builtin}")
-    lines.append(f"  Compiled: {compiled}")
-    lines.append(f"  Regular: {regular}")
-    lines.append(f"  Stubs: {stubs}")
-    lines.append("  Per extension:")
-    for ext, number in sorted(per_ext.items()):
-        if ext:
-            lines.append(f"    {ext}: {number}")
-    visit_time = stats["time_spent_visiting"] / 1000
-    inspect_time = stats["time_spent_inspecting"] / 1000
-    total_time = visit_time + inspect_time
-    visit_percent = visit_time / total_time * 100
-    inspect_percent = inspect_time / total_time * 100
-    try:
-        visit_time_per_module = visit_time / regular
-    except ZeroDivisionError:
-        visit_time_per_module = 0
-    inspected_modules = builtin + compiled
-    try:
-        inspect_time_per_module = visit_time / inspected_modules
-    except ZeroDivisionError:
-        inspect_time_per_module = 0
-    lines.append("")
-    lines.append(
-        f"Time spent visiting modules ({regular}): "
-        f"{visit_time}ms, {visit_time_per_module:.02f}ms/module ({visit_percent:.02f}%)"
-    )
-    lines.append(
-        f"Time spent inspecting modules ({inspected_modules}): "
-        f"{inspect_time}ms, {inspect_time_per_module:.02f}ms/module ({inspect_percent:.02f}%)"
-    )
-    serialize_time = stats["time_spent_serializing"] / 1000
-    serialize_time_per_module = serialize_time / modules
-    lines.append(f"Time spent serializing: " f"{serialize_time}ms, {serialize_time_per_module:.02f}ms/module")
-    return "\n".join(lines)
-
-
 def _load_packages(
     packages: Sequence[str],
     *,
-    extensions: Extensions | None,
-    search_paths: Sequence[str | Path],
-    docstring_parser: Parser | None,
-    docstring_options: dict[str, Any] | None,
+    extensions: Extensions | None = None,
+    search_paths: Sequence[str | Path] | None = None,
+    docstring_parser: Parser | None = None,
+    docstring_options: dict[str, Any] | None = None,
     resolve_aliases: bool = True,
     resolve_implicit: bool = False,
     resolve_external: bool = False,
     allow_inspection: bool = True,
+    store_source: bool = True,
+    find_stubs_package: bool = False,
 ) -> GriffeLoader:
+    # Create a single loader.
     loader = GriffeLoader(
         extensions=extensions,
         search_paths=search_paths,
         docstring_parser=docstring_parser,
         docstring_options=docstring_options,
         allow_inspection=allow_inspection,
+        store_source=store_source,
     )
+
+    # Load each package.
     for package in packages:
         if not package:
             logger.debug("Empty package name, continuing")
             continue
         logger.info(f"Loading package {package}")
         try:
-            loader.load_module(package)
+            loader.load(package, try_relative_path=True, find_stubs_package=find_stubs_package)
         except ModuleNotFoundError as error:
-            logger.error(f"Could not find package {package}: {error}")
+            logger.error(f"Could not find package {package}: {error}")  # noqa: TRY400
         except ImportError as error:
-            logger.error(f"Tried but could not import package {package}: {error}")
+            logger.exception(f"Tried but could not import package {package}: {error}")  # noqa: TRY401
     logger.info("Finished loading packages")
+
+    # Resolve aliases.
     if resolve_aliases:
         logger.info("Starting alias resolution")
         unresolved, iterations = loader.resolve_aliases(implicit=resolve_implicit, external=resolve_external)
@@ -179,13 +114,20 @@ def _load_packages(
 _level_choices = ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL")
 
 
+def _extensions_type(value: str) -> Sequence[str | dict[str, Any]]:
+    try:
+        return json.loads(value)
+    except json.JSONDecodeError:
+        return value.split(",")
+
+
 def get_parser() -> argparse.ArgumentParser:
-    """Return the program argument parser.
+    """Return the CLI argument parser.
 
     Returns:
-        The argument parser for the program.
+        An argparse parser.
     """
-    usage = "%(prog)s [GLOBAL_OPTS...] COMMAND [COMMAND_OPTS...]"  # noqa: WPS323 (%-formatting)
+    usage = "%(prog)s [GLOBAL_OPTS...] COMMAND [COMMAND_OPTS...]"
     description = "Signatures for entire Python programs. "
     "Extract the structure, the frame, the skeleton of your project, "
     "to generate API documentation or find breaking changes in your API."
@@ -196,8 +138,10 @@ def get_parser() -> argparse.ArgumentParser:
 
     global_options = parser.add_argument_group(title="Global options")
     global_options.add_argument("-h", "--help", action="help", help=main_help)
+    global_options.add_argument("-V", "--version", action="version", version=f"%(prog)s {debug.get_version()}")
+    global_options.add_argument("--debug-info", action=_DebugInfo, help="Print debug information.")
 
-    def add_common_options(subparser):  # noqa: WPS430
+    def add_common_options(subparser: argparse.ArgumentParser) -> None:
         common_options = subparser.add_argument_group(title="Common options")
         common_options.add_argument("-h", "--help", action="help", help=subcommand_help)
         search_options = subparser.add_argument_group(title="Search options")
@@ -210,11 +154,19 @@ def get_parser() -> argparse.ArgumentParser:
             help="Paths to search packages into.",
         )
         loading_options = subparser.add_argument_group(title="Loading options")
+        loading_options.add_argument(
+            "-B",
+            "--find-stubs-packages",
+            dest="find_stubs_package",
+            action="store_true",
+            default=False,
+            help="Whether to look for stubs-only packages and merge them with concrete ones.",
+        )
         loading_options.add_argument(
             "-e",
             "--extensions",
             default={},
-            type=json.loads,
+            type=_extensions_type,
             help="A list of extensions to use.",
         )
         loading_options.add_argument(
@@ -238,10 +190,14 @@ def get_parser() -> argparse.ArgumentParser:
 
     # ========= SUBPARSERS ========= #
     subparsers = parser.add_subparsers(
-        dest="subcommand", title="Commands", metavar="COMMAND", prog="griffe", required=True
+        dest="subcommand",
+        title="Commands",
+        metavar="COMMAND",
+        prog="griffe",
+        required=True,
     )
 
-    def add_subparser(command: str, text: str, **kwargs) -> argparse.ArgumentParser:  # noqa: WPS430 (nested function)
+    def add_subparser(command: str, text: str, **kwargs: Any) -> argparse.ArgumentParser:
         return subparsers.add_parser(command, add_help=False, help=text, description=text, **kwargs)
 
     # ========= DUMP PARSER ========= #
@@ -327,7 +283,23 @@ def get_parser() -> argparse.ArgumentParser:
         metavar="BASE_REF",
         help="Git reference (commit, branch, tag) to check. Default: load current code.",
     )
+    check_options.add_argument(
+        "--color",
+        dest="color",
+        action="store_true",
+        default=None,
+        help="Force enable colors in the output.",
+    )
+    check_options.add_argument(
+        "--no-color",
+        dest="color",
+        action="store_false",
+        default=None,
+        help="Force disable colors in the output.",
+    )
     check_options.add_argument("-v", "--verbose", action="store_true", help="Verbose output.")
+    formats = ("oneline", "verbose")
+    check_options.add_argument("-f", "--format", dest="style", choices=formats, default=None, help="Output format.")
     add_common_options(check_parser)
 
     return parser
@@ -340,11 +312,12 @@ def dump(
     full: bool = False,
     docstring_parser: Parser | None = None,
     docstring_options: dict[str, Any] | None = None,
-    extensions: Sequence[str | dict[str, Any] | Extension | Type[Extension]] | None = None,
+    extensions: Sequence[str | dict[str, Any] | ExtensionType | type[ExtensionType]] | None = None,
     resolve_aliases: bool = False,
     resolve_implicit: bool = False,
     resolve_external: bool = False,
     search_paths: Sequence[str | Path] | None = None,
+    find_stubs_package: bool = False,
     append_sys_path: bool = False,
     allow_inspection: bool = True,
     stats: bool = False,
@@ -362,6 +335,9 @@ def dump(
         resolve_external: Whether to load additional, unspecified modules to resolve aliases.
         extensions: The extensions to use.
         search_paths: The paths to search into.
+        find_stubs_package: Whether to search for stubs-only packages.
+            If both the package and its stubs are found, they'll be merged together.
+            If only the stubs are found, they'll be used as the package itself.
         append_sys_path: Whether to append the contents of `sys.path` to the search paths.
         allow_inspection: Whether to allow inspecting modules when visiting them is not possible.
         stats: Whether to compute and log stats about loading.
@@ -369,6 +345,7 @@ def dump(
     Returns:
         `0` for success, `1` for failure.
     """
+    # Prepare options.
     per_package_output = False
     if isinstance(output, str) and output.format(package="package") != output:
         per_package_output = True
@@ -380,9 +357,10 @@ def dump(
     try:
         loaded_extensions = load_extensions(extensions or ())
     except ExtensionError as error:
-        logger.error(error)
+        logger.exception(str(error))  # noqa: TRY401
         return 1
 
+    # Load packages.
     loader = _load_packages(
         packages,
         extensions=loaded_extensions,
@@ -393,10 +371,13 @@ def dump(
         resolve_implicit=resolve_implicit,
         resolve_external=resolve_external,
         allow_inspection=allow_inspection,
+        store_source=False,
+        find_stubs_package=find_stubs_package,
     )
     data_packages = loader.modules_collection.members
 
-    started = datetime.now()
+    # Serialize and dump packages.
+    started = datetime.now(tz=timezone.utc)
     if per_package_output:
         for package_name, data in data_packages.items():
             serialized = data.as_json(indent=2, full=full)
@@ -404,10 +385,10 @@ def dump(
     else:
         serialized = json.dumps(data_packages, cls=JSONEncoder, indent=2, full=full)
         _print_data(serialized, output)
-    elapsed = datetime.now() - started
+    elapsed = datetime.now(tz=timezone.utc) - started
 
     if stats:
-        logger.info(_stats({"time_spent_serializing": elapsed.microseconds, **loader.stats()}))
+        logger.info(_format_stats({"time_spent_serializing": elapsed.microseconds, **loader.stats()}))
 
     return 0 if len(data_packages) == len(packages) else 1
 
@@ -418,10 +399,13 @@ def check(
     against_path: str | Path | None = None,
     *,
     base_ref: str | None = None,
-    extensions: Sequence[str | dict[str, Any] | Extension | Type[Extension]] | None = None,
+    extensions: Sequence[str | dict[str, Any] | ExtensionType | type[ExtensionType]] | None = None,
     search_paths: Sequence[str | Path] | None = None,
+    find_stubs_package: bool = False,
     allow_inspection: bool = True,
     verbose: bool = False,
+    color: bool | None = None,
+    style: str | ExplanationStyle | None = None,
 ) -> int:
     """Load packages data and dump it as JSON.
 
@@ -438,21 +422,24 @@ def check(
     Returns:
         `0` for success, `1` for failure.
     """
-    colorama.deinit()
-    colorama.init()
-
+    # Prepare options.
     search_paths = list(search_paths) if search_paths else []
 
-    against = against or _get_latest_tag(package)
+    try:
+        against = against or _get_latest_tag(package)
+    except GitError as error:
+        print(f"griffe: error: {error}", file=sys.stderr)
+        return 2
     against_path = against_path or package
     repository = _get_repo_root(against_path)
 
     try:
         loaded_extensions = load_extensions(extensions or ())
     except ExtensionError as error:
-        logger.error(error)
+        logger.exception(str(error))  # noqa: TRY401
         return 1
 
+    # Load old and new version of the package.
     old_package = load_git(
         against_path,
         ref=against,
@@ -469,6 +456,7 @@ def check(
             extensions=loaded_extensions,
             search_paths=search_paths,
             allow_inspection=allow_inspection,
+            find_stubs_package=find_stubs_package,
         )
     else:
         new_package = load(
@@ -477,21 +465,27 @@ def check(
             extensions=loaded_extensions,
             search_paths=search_paths,
             allow_inspection=allow_inspection,
+            find_stubs_package=find_stubs_package,
         )
 
-    if verbose:
-        style = ExplanationStyle.VERBOSE
-    else:
-        style = ExplanationStyle.ONE_LINE
+    # Find and display API breakages.
     breakages = list(find_breaking_changes(old_package, new_package))
+
+    colorama.deinit()
+    colorama.init(strip=color if color is None else not color)
+    if style is None:
+        style = ExplanationStyle.VERBOSE if verbose else ExplanationStyle.ONE_LINE
+    else:
+        style = ExplanationStyle(style)
     for breakage in breakages:
         print(breakage.explain(style=style), file=sys.stderr)
+
     if breakages:
         return 1
     return 0
 
 
-def main(args: list[str] | None = None) -> int:  # noqa: WPS231
+def main(args: list[str] | None = None) -> int:
     """Run the main program.
 
     This function is executed when you type `griffe` or `python -m griffe`.
@@ -502,11 +496,14 @@ def main(args: list[str] | None = None) -> int:  # noqa: WPS231
     Returns:
         An exit code.
     """
+    # Parse arguments.
     parser = get_parser()
     opts: argparse.Namespace = parser.parse_args(args)
     opts_dict = opts.__dict__
+    opts_dict.pop("debug_info")
     subcommand = opts_dict.pop("subcommand")
 
+    # Initialize logging.
     log_level = opts_dict.pop("log_level", DEFAULT_LOG_LEVEL)
     try:
         level = getattr(logging, log_level)
@@ -518,7 +515,11 @@ def main(args: list[str] | None = None) -> int:  # noqa: WPS231
         )
         return 1
     else:
-        logging.basicConfig(format="%(levelname)-10s %(message)s", level=level)  # noqa: WPS323
+        logging.basicConfig(format="%(levelname)-10s %(message)s", level=level)
 
+    # Run subcommand.
     commands: dict[str, Callable[..., int]] = {"check": check, "dump": dump}
     return commands[subcommand](**opts_dict)
+
+
+__all__ = ["check", "dump", "get_parser", "main"]
