@@ -4,17 +4,23 @@ from __future__ import annotations
 
 from ast import PyCF_ONLY_AST
 from contextlib import suppress
-from typing import TYPE_CHECKING, Callable
+from typing import TYPE_CHECKING, Protocol
 
-from griffe.agents.nodes import get_annotation
-from griffe.expressions import Expression, Name
-from griffe.logger import get_logger
+from griffe.agents.nodes import safe_get_annotation
+from griffe.exceptions import BuiltinModuleError
+from griffe.logger import LogLevel, get_logger
 
 if TYPE_CHECKING:
     from griffe.dataclasses import Docstring
+    from griffe.expressions import Expr
 
 
-def warning(name: str) -> Callable[[Docstring, int, str], None]:
+class WarningCallable(Protocol):
+    def __call__(self, docstring: Docstring, offset: int, message: str, log_level: LogLevel = ...) -> None:
+        ...
+
+
+def warning(name: str) -> WarningCallable:
     """Create and return a warn function.
 
     Parameters:
@@ -32,33 +38,48 @@ def warning(name: str) -> Callable[[Docstring, int, str], None]:
     """
     logger = get_logger(name)
 
-    def warn(docstring: Docstring, offset: int, message: str) -> None:  # noqa: WPS430
+    def warn(docstring: Docstring, offset: int, message: str, log_level: LogLevel = LogLevel.warning) -> None:
         try:
             prefix = docstring.parent.relative_filepath  # type: ignore[union-attr]
         except (AttributeError, ValueError):
             prefix = "<module>"
-        logger.warning(f"{prefix}:{(docstring.lineno or 0)+offset}: {message}")
+        except BuiltinModuleError:
+            prefix = f"<module: {docstring.parent.module.name}>"  # type: ignore[union-attr]
+        log = getattr(logger, log_level.value)
+        log(f"{prefix}:{(docstring.lineno or 0)+offset}: {message}")
 
     return warn
 
 
-def parse_annotation(annotation: str, docstring: Docstring) -> str | Name | Expression:
+def parse_annotation(
+    annotation: str,
+    docstring: Docstring,
+    log_level: LogLevel = LogLevel.error,
+) -> str | Expr:
     """Parse a string into a true name or expression that can be resolved later.
 
     Parameters:
         annotation: The annotation to parse.
         docstring: The docstring in which the annotation appears.
             The docstring's parent is accessed to bind a resolver to the resulting name/expression.
+        log_level: Log level to use to log a message.
 
     Returns:
         The string unchanged, or a new name or expression.
     """
     with suppress(
         AttributeError,  # docstring has no parent that can be used to resolve names
-        KeyError,  # compiled annotation contains unhandled AST nodes
         SyntaxError,  # annotation contains syntax errors
     ):
         code = compile(annotation, mode="eval", filename="", flags=PyCF_ONLY_AST, optimize=2)
-        if code.body:
-            return get_annotation(code.body, parent=docstring.parent) or annotation  # type: ignore[arg-type]
+        if code.body:  # type: ignore[attr-defined]
+            name_or_expr = safe_get_annotation(
+                code.body,  # type: ignore[attr-defined]
+                parent=docstring.parent,
+                log_level=log_level,
+            )
+            return name_or_expr or annotation
     return annotation
+
+
+__all__ = ["parse_annotation", "warning"]
