@@ -2,26 +2,42 @@
 
 from __future__ import annotations
 
-from typing import Any, Callable, List, Tuple, Union
+from typing import TYPE_CHECKING, Any, Iterator, List, Protocol, Tuple, Union
 
 from griffe.dataclasses import Attribute, Class, Docstring, Function, Module
 from griffe.docstrings.dataclasses import DocstringAttribute, DocstringElement, DocstringParameter, DocstringSection
+from griffe.logger import LogLevel
+
+if TYPE_CHECKING:
+    from types import ModuleType
+
 
 ParentType = Union[Module, Class, Function, Attribute, None]
 ParseResultType = Tuple[List[DocstringSection], List[str]]
 
 
-def parser(parser_module) -> Callable[[str, ParentType, Any], ParseResultType]:
+class ParserType(Protocol):  # noqa: D101
+    def __call__(  # noqa: D102
+        self,
+        docstring: str,
+        parent: ParentType | None = None,
+        **parser_opts: Any,
+    ) -> ParseResultType:
+        ...
+
+
+def parser(parser_module: ModuleType) -> Iterator[ParserType]:
     """Wrap a parser to help testing.
 
     Parameters:
         parser_module: The parser module containing a `parse` function.
 
-    Returns:
+    Yields:
         The wrapped function.
     """
+    original_warn = parser_module._warn
 
-    def parse(docstring: str, parent: ParentType = None, **parser_opts: Any) -> ParseResultType:  # noqa: WPS430
+    def parse(docstring: str, parent: ParentType | None = None, **parser_opts: Any) -> ParseResultType:
         """Parse a doctring.
 
         Parameters:
@@ -38,11 +54,13 @@ def parser(parser_module) -> Callable[[str, ParentType, Any], ParseResultType]:
             docstring_object.parent = parent
             parent.docstring = docstring_object
         warnings = []
-        parser_module._warn = lambda _docstring, _offset, message: warnings.append(message)  # noqa: WPS437
+        parser_module._warn = lambda _docstring, _offset, message, log_level=LogLevel.warning: warnings.append(message)  # type: ignore[attr-defined]
         sections = parser_module.parse(docstring_object, **parser_opts)
         return sections, warnings
 
-    return parse  # type: ignore
+    yield parse
+
+    parser_module._warn = original_warn  # type: ignore[attr-defined]
 
 
 def assert_parameter_equal(actual: DocstringParameter, expected: DocstringParameter) -> None:
@@ -75,5 +93,10 @@ def assert_element_equal(actual: DocstringElement, expected: DocstringElement) -
         actual: The actual element.
         expected: The expected element.
     """
-    assert actual.annotation == expected.annotation  # type: ignore[operator]
-    assert actual.description == expected.description
+    assert isinstance(actual, type(expected))
+
+    for k in actual.as_dict():
+        src = getattr(actual, k)
+        dst = getattr(expected, k)
+
+        assert src == dst, f"attribute {k!r}, {src!r} != {dst!r}"
